# Auditing:

The following document will provide the steps required to enable auditing. The resulting configuration will have a deployment of MindLink Anywhere with a logger that provides the audit events and a script to aggregate audited group events. The output files will be provided in a single folder for auditors to analyze.

We recommend the following practices:
- Set up the serilog configuration to put audit json files into an audit directory
- Set up the audit script to periodically scrape into the same audit directory
- Have the directory monitored for audit content

## Configuring the logger via the Serilog configuration file

While running the service it is possible to configure the logger directly via a `<install directory>\Config\serilog.json` file.

This is what the default file looks like to enable a file logger for verbose messages and event log for errors:

```json
{
    "Serilog":{
        "MinimumLevel": "Verbose",
        "Using": ["MindLink.Core.Common", "Serilog.Sinks.File", "Serilog.Sinks.EventLog" ],
        "WriteTo:Main": {
            "Name": "Logger",
            "Args": {
                "configureLogger": {
                    "WriteTo": [
                    {
                        "Name": "File",
                        "Args": {
                            "path": "<install directory>\\logs\\Connector.log",
                            "rollingInterval": "Day",
                            "fileSizeLimitBytes": "100000000",
                            "rollOnFileSizeLimit": "true",
                            "outputTemplate": "[{Level}] {Timestamp:yyyy-MM-ddTHH:mm:ss.fffZ} {MachineName} [{SourceContext}] {Message:lj}{NewLine}{Exception}",
                            "useMindLinkFormatter": "true"
                        }
                    }]
                }
            }
        },
        "WriteTo:Event": {
            "Name": "EventLog",
            "Args": {
                "source": "Connector Server",
                "restrictedToMinimumLevel": "Error"
            }
        }
    }
}
```
Some changes to the configuration file can be made while the service is running. The following relevant properties can be updated dynamically:
 - Logging Levels
 - Logging Level Overrides (existing overrides only)

Other changes will require a host restart before taking effect.

### Logging levels and "MinimumLevel"

Serilog separates logging into muliple levels:
 - Verbose:
    - The noisiest logging level.
 - Debug:
    - Typically used for internal system events that are not usually viewable from the outside.
 - Warning:
    - When the service is endangered or may be behaving outside its expected parameters.
 - Error:
    - When functionality is unavailable or expectations broken.
 - Fatal:
    - The most critical level.

By setting the **MinimumLevel** of the logger you can restrict or reveal more information provided by the logger.

### Enabling Auditing
The following sample demonstrates how to add a sub-logger for auditing.
```json
{
    "Serilog":{
        "MinimumLevel": "Verbose",
        "Using": ["MindLink.Core.Common", "Serilog.Sinks.File", "Serilog.Sinks.EventLog" ],
        "WriteTo:Main": {
            "Name": "Logger",
            "Args": {
                "configureLogger": {
                    "WriteTo": [
                    {
                        "Name": "File",
                        "Args": {
                            "path": "<install directory>\\logs\\Connector.log",
                            "rollingInterval": "Day",
                            "fileSizeLimitBytes": "100000000",
                            "rollOnFileSizeLimit": "true",
                            "outputTemplate": "[{Level}] {Timestamp:yyyy-MM-ddTHH:mm:ss.fffZ} {MachineName} [{SourceContext}] {Message:lj}{NewLine}{Exception}",
                            "useMindLinkFormatter": "true"
                        }
                    }]
                }
            }
        },
        "WriteTo:Audit": {
            "Name": "Logger",
            "Args": {
                "configureLogger": {
                    "Filter": [
                        {
                            "Name": "ByIncludingOnly",
                            "Args": {
                                "expression": "Has(AuditEntry)"
                            }
                        }
                    ],
                    "WriteTo": [
                    {
                        "Name": "File",
                        "Args": {
                            "path": "%PROGRAMDATA%\\MindLink\\audit.json",
                            "rollingInterval": "Day",
                            "formatter": "Serilog.Formatting.Compact.CompactJsonFormatter,Serilog.Formatting.Compact",
                            "fileSizeLimitBytes": "100000000",
                            "rollOnFileSizeLimit": "true"
                        }
                    }]
                }
            }
        },
        "WriteTo:Event": {
            "Name": "EventLog",
            "Args": {
                "source": "Connector Server",
                "restrictedToMinimumLevel": "Error"
            }
        }
    }
}

```
Note that the output of the Audit sub-logger can be configured to include certain types of Audit events please refer to the AuditEntry types listed in the Logging Page on our documentation site: https://docs.mindlinksoft.com/docs/Manage_And_Maintain/Logging/Logging#audit-entry-types

The file type of the audit log can be configured by changing the file extension  of the file. e.g. "path": "%PROGRAMDATA%\\MindLink\\audit.log" -> "path": "%PROGRAMDATA%\\MindLink\\audit.json"

## How to capture MCE group events

### Description
Included in the PowerShell module are three commandlets:
- Update-DatabaseForAuditing
   - Updates event tables to include a unique ID column to the event tables to be audited.
- Get-AuditEvents
   - Produces a list of objects, each representing an audit event from the database.
- Save-AuditEvents
   - Takes in a list of objects representing the audit events to save into a particular file.


Running all functions correctly will produce a singular _json_ file containing audit records for when a user:
- Gains read rights to a group
- Loses read rights to a group
- Joins a group
- Leaves a group
- Sends a message

If not specified, this filename of the results will contain the unix timestamp of the time at which running all functions was executed.
If attempting to save results with a duplicate filename, numbers will be appended to ensure uniqueness and that no file is overridden.

The records will be in chronological order specified by the timestamp of when the event occurred.
- Each record will contain the following information:
   - The type of the event
   - The time of the event
      - A timestamp as UNIX time
      - A human readable UTC timestring
   - The ID of the group
   - The name of the group
   - The security context of the group
   - The ID of the user
   - The name of the user
   - The linked identity of the user

If the event concerns a sent message, the message content will __not__ be included, instead the message ID will be included for further auditing purposes.

### How to run
It is not a requirement that the SQL script and PowerShell script be kept within the same working directory as full file paths can be specified for both the SQL script and output json file.

Before being able to access the inner functions, __it is required that you import__ the _GenerateAuditEvents.ps1_ script with:
`Import-Module ./GenerateAuditEvents.ps1`

When using this script for the first time in order to audit any events, _Update-DatabaseForAuditing_ __must be used to update the database.__
If not executed, you may come across an error messaging mentioning "Invalid column name 'GlobalEventId'". This column is required to fetch all the events from the relevent tables.

After successfully running the _Get-AuditEvents_ cmdlet, a "LastEvents.json" file will be generated in the working directory of the script.
This file contains cursor information about the previous execution, which is used in subsequent executions. This will retrieve the new events that have occurred since the last execution.
Deleting this file will cause the next run to retrieve all events.

Updating the database with __Update-DatabaseForAuditing__ specifies the following parameters:

- __DatabaseConnectionString:__
Mandatory parameter used to determine the connection string to connect to the MCE database.

- __SqlScriptPath:__
Mandatory parameter used to specify the path to the SQL script.
The script, "PrimeDbForAuditing.sql", should have been included packaged with this PowerShell script.

- __UpdateType:__
Mandatory parameter used to specify how you are updating the table columns. 
Values can be:
 - Add: case-insensitive, will add a unique id column in the tables of the events.
 - Remove: case-insensitive, will remove a unique id column in the tables of the events.

___Steps to prime the database for auditing:___
- Open PowerShell instance and navigate to script directory
- `Import-Module ./GenerateAuditEvents.ps1`
- Run _Update-DatabaseForAuditing_

  ```powershell
  PS> Update-DatabaseForAuditing -DatabaseConnectionString 'Server=.;Initial Catalog=Mce;Integrated Security=SSPI' -SqlScriptPath "PrimeDbForAuditing.sql" -UpdateType "Add"
  ```

___Steps to reset the database from auditing changes:___
- Open PowerShell instance and navigate to script directory
- `Import-Module ./GenerateAuditEvents.ps1`</code>`
- Run _Update-DatabaseForAuditing_
  ```powershell
  PS> Update-DatabaseForAuditing -DatabaseConnectionString 'Server=.;Initial Catalog=Mce;Integrated Security=SSPI' -SqlScriptPath "PrimeDbForAuditing.sql" -UpdateType "Remove"
  ```

Generating audit events with __Get-AuditEvents__ specifies the following parameters:

- __DatabaseConnectionString:__
Mandatory parameter used to determine the connection string to connect to the MCE database. 

- __SqlScriptPath:__
Mandatory parameter used to specify the path to the SQL script.
The script, "GenerateAuditEvents.sql", should have been included packaged with this PowerShell script.

- __Timeout:__
Optional parameter used to specify the wait time (in seconds) before terminating the attempt to execute the sql script and generating an error.
If not specified then the parameter will default 0, indicating no timeout limit.

___Steps:___
- Open PowerShell instance and navigate to script directory
- `Import-Module ./GenerateAuditEvents.ps1`
- Run _Get-AuditEvents_ using the parameters specified above (use Get-Help for more information on the commandlet)
  ```powershell
  Get-AuditEvents -DatabaseConnectionString {MyDatabaseConnectionString} -SqlScriptPath {GenerateAuditEvents.sql}
  ```

Saving audit events with __Save-AuditEvents__ specifies the following parameters:

- __AuditEvents:__
Mandatory parameter supplied by the pipeline, this is an ordered list of System.Object each representing an audit event entry.

- __OutFile:__
Optional parameter which specifies the output file which contains the generated audit events.
If not specified, will default to "AuditResults{CurrentTimestamp}.json" saving a file within the running directory, where the timestamp is derived from the time that the script runs.

___Steps:___
- Open PowerShell instance and navigate to script directory
- `Import-Module ./GenerateAuditEvents.ps1`
- Run _Save-AuditEvents_ using the output from _Get-AuditEvents_

   ```powershell
   Get-AuditEvents -DatabaseConnectionString {MyDatabaseConnectionString} -SqlScriptPath {GenerateAuditEvents.sql} | Save-AuditEvents -OutFile {MyCustomFileName}
   ```

---

### Examples
- _Running against a localhost for the first time, specifying the full path of the SQL script and saving the json file to the working directory_
  ```powershell
  PS> Update-DatabaseForAuditing -DatabaseConnectionString 'Server=.;Initial Catalog=Mce;Integrated Security=SSPI' -SqlScriptPath "PrimeDbForAuditing.sql" -UpdateType "Add"
  PS> Get-AuditEvents -DatabaseConnectionString 'Server=localhost;Initial Catalog=MCE;Integrated Security=SSPI' -SqlScriptPath 'GenerateAuditEvents.sql' | Save-AuditEvents
  ```

- _Running against a DB with credentials, specifying the name of the SQL script within the same working directory and saving the json file to a separate directory_

  ```powershell
  PS> .\GenerateAuditEvents.ps1 -DatabaseConnectionString 'Server=;Initial Catalog=mce-test;User Id=myUsername;Password=myPassword;' -SqlScriptPath 'C:\eng\GenerateAuditEvents.sql' | Save-AuditEvents -OutFile 'C:\files\MyCustomFileName'
  ```


## How to periodically scrape the database for audit events

Requirements:
- Windows Task Scheduler
- Windows Powershell
- A MCE Database that has been primed for auditing. (See How to capture MCE group events - How To Run)

While the examples above demonstrate how to run these powershell functions manually it is possible to automate this process via Windows Task Scheduler.

### Creating a scheduled Task via Task Scheduler

1. Open Task Scheduler by searching for "Task Scheduler" via the windows search bar or pressing "Windows key + R" and entering "taskschd.msc".

2. Click "Create a task" in the Actions section of Task Scheduler.

3. Enter a memorable name for the name of the task. 

4. Switch to the "triggers" tab and click the "new" button.

5. You can configure how often the task wil be run here.

6. After creating the trigger navigate to the Actions tab and click "new". 

7. Enter "powershell" as the program.

8. To provide the task with the script enter the following as the arguments:

   ```powershell
   -File "InvokeGenerateAuditEvents.ps1" -DatabaseConnectionString "Your database connection string" -SqlScriptPath "your directory for the SQL path, or the name of the file if in the same folder" -OutFile "Your audit directory"
   ``` 
   - 8a. Make sure that you do not use single quotation marks.

9. Enter the path to the directory where you keep the audit script in the 'Start in' field.

The "Conditions"  and "Settings" tabs provide additional configuration. Please use them according to your needs. 

### Creating a scheduled task via power shell

Alternatively, you can create a scheduled task with the following:

```powershell
$Trigger= New-ScheduledTaskTrigger -At 10:00am –Daily  # Specify the trigger settings

$Action= New-ScheduledTaskAction -Execute "PowerShell.exe" -Argument '-File "InvokeGenerateAuditEvents.ps1" -DatabaseConnectionString "Your database connection string" -SqlScriptPath "your path for the SQL script, or the name of the file if it is in the same folder" -OutFile "Your audit directory"' –WorkingDirectory "The path to the directory where the powershell script is located" #Specify what program to run with its parameters

Register-ScheduledTask -TaskName "Audit MCE group events" -Trigger $Trigger -Action $Action -RunLevel Highest –Force  # Specify the name and register the task
```