<#
.Description
Queries the MCE database and outputs relavent audit events to the command line for when a user:
- Gains read rights to a group
- Loses read rights to a group
- Joins a group
- Leaves a group
- Sends a message

Upon running the script a "LastEvent.txt" file will be generated in the working directory of the script.
This file contains cursor information about the previous execution, which is used in subsequent executions to retrieve the new events since the last execution.

.PARAMETER DatabaseConnectionString
Mandatory parameter used to determine the connection string to connect to the MCE database. 

.PARAMETER SqlScriptPath
Mandatory parameter used to specify the path to the SQL script.
The script should have been included packaged with this PowerShell script.

.PARAMETER Timeout
Optional parameter used to specify the wait time (in seconds) before terminating the attempt to execute the sql script and generating an error.
If not specified then the parameter will default 0, indicating no timeout limit.

.EXAMPLE
PS> .\GenerateAuditEvents.ps1 -DatabaseConnectionString 'Server=.;Initial Catalog=MCE;Integrated Security=SSPI' -SqlScriptPath 'C:\eng\GenerateAuditEvents.sql'
.EXAMPLE
PS> .\GenerateAuditEvents.ps1 -DatabaseConnectionString 'Server=;Initial Catalog=mce-test;User Id=myUsername;Password=myPassword;' -SqlScriptPath 'GenerateAuditEvents.sql' -Timeout 300

.SYNOPSIS
Generates relavent audit events from the given database.
#>

function Get-AuditEvents {
# Parameters
param (
    [string] [Parameter(Mandatory)] $DatabaseConnectionString,
    [string] [Parameter(Mandatory)] $SqlScriptPath,
    [int] $Timeout=0);

    # Constants.
    $LastEventReceivedFileName = "LastEvents.json";
    $LastGroupRosterEventId = 0;
    $LastGroupMessageEventId = 0;
    $LastGroupUserRightEventId = 0;

    # Check if last event ids exists.
    if (Test-Path -Path $LastEventReceivedFileName -PathType Leaf) {
        $LastEvents = Get-Content -Path $LastEventReceivedFileName -Raw | ConvertFrom-Json;

        $LastGroupRosterEventId = $LastEvents.LatestGlobalEventId[0];
        $LastGroupUserRightEventId = $LastEvents.LatestGlobalEventId[1];
        $LastGroupMessageEventId = $LastEvents.LatestGlobalEventId[2];
    }

    # Read in SQL query
    $Query = Get-Content -Path $SqlScriptPath -Raw;

    # Setup database connection using connection string.
    $connection = New-Object System.Data.SqlClient.SqlConnection;
    $connection.ConnectionString = $DatabaseConnectionString;
    $connection.Open();

    # Setup SQL command.
    $command = $connection.CreateCommand();

    $command.CommandText = $query;
    $command.CommandTimeout = $CommandTimeout;

    $command.Parameters.Add((New-OBJECT DATA.SQLClient.SQLParameter("@LastGroupMessagesGlobalEventId",[Data.SQLDBType]::BigInt)));
    $command.Parameters.Add((New-OBJECT DATA.SQLClient.SQLParameter("@LastGroupUserRightsGlobalEventId",[Data.SQLDBType]::BigInt)));
    $command.Parameters.Add((New-OBJECT DATA.SQLClient.SQLParameter("@LastGroupEndpointRosterGlobalEventId",[Data.SQLDBType]::BigInt)));

    $command.Parameters["@LastGroupMessagesGlobalEventId"].Value = $LastGroupMessageEventId;
    $command.Parameters["@LastGroupUserRightsGlobalEventId"].Value = $LastGroupUserRightEventId;
    $command.Parameters["@LastGroupEndpointRosterGlobalEventId"].Value = $LastGroupRosterEventId;

    $adapter = New-Object System.Data.sqlclient.sqlDataAdapter $command;
    $dataset = New-Object System.Data.DataSet;

    # Execute command
    try {
        $adapter.Fill($dataset)
    }
    catch
    {
        Write-Error $_.Exception -ErrorAction Stop;
    }

    $resultTable = $dataset.Tables[0];

    # Clean up connection.
    $connection.Close();
    
    # Exclude non-relavent properties.
    $obj = $resultTable | Select-Object * -ExcludeProperty ItemArray, Table, RowError, RowState, HasErrors;

    # Iterate through properties to remove null values, format the security context JSON data and format the timestamp and timestrings.
    foreach ($element in $obj) {
        if ($element.psobject.Properties["MessageId"].value -eq [DBNull]::Value) {
            $element.psobject.Properties.Remove("MessageId");
        }

        $element.psobject.Properties["Timestamp"].value = $element.psobject.Properties["Timestamp"].value.ToUnixTimeMilliseconds();
        $element.psobject.Properties["Timestring"].value = $element.psobject.Properties["Timestring"].value.ToUniversalTime().ToString("o");

        $temp = $element.psobject.Properties["GroupSecurityContexts"].value | ConvertFrom-Json;
        $element.psobject.Properties["GroupSecurityContexts"].value = $temp;
    }

    # Write out metadata from successful run.
    $tableMetadata = $dataset.Tables[1] | Select-Object * -ExcludeProperty ItemArray, Table, RowError, RowState, HasErrors;
    ConvertTo-Json -InputObject $tableMetadata | Out-File $LastEventReceivedFileName;

    Write-Output -NoEnumerate $obj;

}

<#
.Description
Alters the tables to be audited to include a unique ID.

.PARAMETER DatabaseConnectionString
Mandatory parameter used to determine the connection string to connect to the MCE database. 

.PARAMETER SqlScriptPath
Mandatory parameter used to specify the path to the SQL script to alter the tables.
The script should have been included packaged with this PowerShell script.

.PARAMETER UpdateType
Mandatory parameter used to specify the path to the SQL script to alter the tables.
Value can either be "Add" or "Remove", case insensitive. 
"Add" will add an additional ID column to the event tables.
"Remove" will remove the additional ID column from the event tables.

.EXAMPLE
PS> Update-DatabaseForAuditing -DatabaseConnectionString 'Server=.;Initial Catalog=Mce;Integrated Security=SSPI' -SqlScriptPath "PrimeDbForAuditing.sql" -UpdateType "Add"
.EXAMPLE
PS> Update-DatabaseForAuditing -DatabaseConnectionString 'Server=.;Initial Catalog=Mce;Integrated Security=SSPI' -SqlScriptPath "PrimeDbForAuditing.sql" -UpdateType "Remove"

.SYNOPSIS
Updates the MCE database for auditing.
#>
function Update-DatabaseForAuditing {
    param(
        [string] [Parameter(Mandatory)] $DatabaseConnectionString,
        [string] [Parameter(Mandatory)] $SqlScriptPath,
        [ValidateSet("Add", "Remove")][string] [Parameter(Mandatory)] $UpdateType
    );

    $UpdateType = $UpdateType.ToLower();

    # Read in SQL query
    $Query = Get-Content -Path $SqlScriptPath -Raw;

    # Setup database connection using connection string.
    $connection = New-Object System.Data.SqlClient.SqlConnection;
    $connection.ConnectionString = $DatabaseConnectionString;
    $connection.Open();

    # Execute the query through a command.
    $command = $connection.CreateCommand();
    $command.CommandText = $query;

    $command.Parameters.ADD((New-OBJECT DATA.SQLClient.SQLParameter("@UpdateType",[Data.SQLDBType]::VarChar, 128))) | OUT-NULL;
    $command.Parameters["@UpdateType"].Value = $UpdateType;

    $command.ExecuteReader();

    # Clean up connection.
    $connection.Close();

}

<#
.Description
Saves the supplied audit events to a specified Json file.
Is intended to have output of `Get-AuditEvents` piped in to function.

.PARAMETER AuditEvents
Mandatory parameter supplied by the pipeline, this is an ordered list of System.Object each representing an audit event entry.

.PARAMETER OutFile
Optional parameter which specifies the output file which contains the generated audit events.
If not specified, will default to "AuditResults{CurrentDateTimestamp}.json" saving a file within the running directory, where the timestamps are derived from the time the script is run.

.EXAMPLE
PS> Get-AuditEvents -DatabaseConnectionString 'Server=.;Initial Catalog=MCE;Integrated Security=SSPI' -SqlScriptPath 'GenerateAuditEvents.sql' | Save-AuditEvents
.EXAMPLE
PS> Get-AuditEvents -DatabaseConnectionString 'Server=.;Initial Catalog=MCE;Integrated Security=SSPI' -SqlScriptPath 'GenerateAuditEvents.sql' | Save-AuditEvents -OutFile "C:\audit\files\CustomFileName.json"

.SYNOPSIS
Saves Json file with supplied audit events.
#>

function Save-AuditEvents {
    # Parameters
    param (
        [System.Object[]] [Parameter(ValueFromPipeline)] $AuditEvents,
        [string] $OutFile
    );

    # Remove '.json' extension if exists
    if ($OutFile -Match '.json$') {
        $OutFile = $OutFile.substring(0, $OutFile.length - 5);
    }

    $currentTime = [DateTimeOffset]::Now.ToUnixTimeSeconds();

    # Default outfile if not supplied.
    if ($OutFile -eq '') {
        $FileName = 'AuditResult' + $currentTime;
    } else {
        $FileName = $OutFile;
    }

    # Check if file exists, if it does we will append numbers to the filename
    $DestFile = $FileName;

    If (Test-Path "$DestFile.json") {
        $i = 1
            While (Test-Path "$DestFile.json") {
                $DestFile = $FileName + '_' + $i
                $i += 1
        }
    }

    $DestFile = "$DestFile.json";

    if( $null -eq $AuditEvents) { 
        ConvertTo-Json -InputObject @() | Out-File -FilePath $DestFile;
    } else {
        # Output the object to desired JSON file.
        $AuditEvents | ConvertTo-Json -Depth 3 | Out-File -FilePath $DestFile;
    }
}